#include "framebuffer.h"
#include "error.h"
#include "gl.h"
#include "shader.h"
#include <stdlib.h>

static struct sw_shaderprogram program_copy;

static char *vert_copy = "#version 300 es\n\
precision highp float;\
in vec2 pos;\
out vec2 out_tex_coord;\
void main() {\
	gl_Position = vec4(pos, 0.0f, 1.0f);\
	out_tex_coord = vec2(pos.x * 0.5 + 0.5, pos.y * 0.5 + 0.5);\
}";

static char *frag_copy = "#version 300 es\n\
precision highp float;\
in vec2 out_tex_coord;\
uniform sampler2D tex;\
out vec4 color;\
void main() {\
	vec4 col = texture(tex, out_tex_coord);\
if(col.a > 0.0) {\
	color = col;\
}\
}";

static f32 screen_rect[12] = {
    -1.f, -1.f, 1.f, -1.f, -1.f, 1.f, 1.f, -1.f, -1.f, 1.f, 1.f, 1.f};
static u32 vao, vbo;

struct sw_framebuffer *sw_framebuffer_create(struct sw_vec2i size) {
	struct sw_framebuffer *fb;

	sw_log("init framebuffer");
	fb = malloc(sizeof(struct sw_framebuffer));
	glGenFramebuffers(1, &fb->_fb);
	glGenTextures(1, &fb->_fbtex);
	glGenTextures(1, &fb->_fbdepth);
	sw_framebuffer_resize(fb, size);

	return fb;
}

void sw_framebuffer_destroy(struct sw_framebuffer *fb) {
	glDeleteTextures(1, &fb->_fbtex);
	glDeleteFramebuffers(1, &fb->_fb);
}

void sw_framebuffer_static_init() {
	sw_log("init static framebuffer");

	program_copy = sw_shaderprogram_create(vert_copy, frag_copy);

	sw_log("init static framebuffer vao");
	glGenVertexArrays(1, &vao);
	glBindVertexArray(vao);

	sw_log("init static framebuffer vbo");
	glGenBuffers(1, &vbo);
	glBindBuffer(GL_ARRAY_BUFFER, vbo);
	glBufferData(
	    GL_ARRAY_BUFFER, sizeof(f32) * 12, screen_rect, GL_STATIC_DRAW
	);
	glGetAttribLocation(program_copy._program, "pos");
	glVertexAttribPointer(
	    glGetAttribLocation(program_copy._program, "pos"),
	    2,
	    GL_FLOAT,
	    GL_FALSE,
	    0,
	    0
	);

	glEnableVertexAttribArray(glGetAttribLocation(program_copy._program, "pos")
	);

	glBindVertexArray(0);
	sw_log("init static framebuffer done");
}

void sw_framebuffer_copy_to_screen(struct sw_framebuffer *fb) {
	glBindVertexArray(vao);
	sw_shaderprogram_use(program_copy);
	glBindTexture(GL_TEXTURE_2D, fb->_fbtex);
	glBindFramebuffer(GL_FRAMEBUFFER, 0);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glDrawArrays(GL_TRIANGLES, 0, 12);
	glBindVertexArray(0);
}

void sw_framebuffer_resize(
    struct sw_framebuffer *fb, struct sw_vec2i new_size
) {
	sw_log("resize framebuffer to %ix%i", new_size.x, new_size.y);
	fb->size = new_size;
	glBindFramebuffer(GL_FRAMEBUFFER, fb->_fb);
	glBindTexture(GL_TEXTURE_2D, fb->_fbtex);
	glTexImage2D(
	    GL_TEXTURE_2D,
	    0,
	    GL_RGBA8,
	    new_size.x,
	    new_size.y,
	    0,
	    GL_RGBA,
	    GL_UNSIGNED_BYTE,
	    0
	);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glFramebufferTexture(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, fb->_fbtex, 0);

	glBindTexture(GL_TEXTURE_2D, fb->_fbdepth);
	glTexImage2D(
	    GL_TEXTURE_2D,
	    0,
	    GL_DEPTH_COMPONENT,
	    new_size.x,
	    new_size.y,
	    0,
	    GL_DEPTH_COMPONENT,
	    GL_UNSIGNED_SHORT,
	    0
	);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glFramebufferTexture(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, fb->_fbdepth, 0);
}
