#include "window.h"

#include <stdlib.h>

#include "error.h"
#include "framebuffer.h"
#include "gl.h"

static void _tick();
static struct sw_window *_tick_window;
static void (*_tick_fn)();
static void GLAPIENTRY gldebug(
    GLenum source,
    GLenum type,
    GLuint id,
    GLenum severity,
    GLsizei length,
    const char *msg,
    const void *user
);

struct sw_window *sw_window_create(struct sw_vec2i size, char *title) {
	struct sw_window *win;

	win = malloc(sizeof(struct sw_window));
	win->size = size;

	sw_log("init glfw");
	glfwInit();
	glfwDefaultWindowHints();

	win->_window = glfwCreateWindow(size.x, size.y, title, 0, 0);
	glfwShowWindow(win->_window);
	glfwMakeContextCurrent(win->_window);
	glewInit();
	glEnable(GL_DEBUG_OUTPUT);
	glDebugMessageCallback(gldebug, 0);
	glfwSwapInterval(1);
	sw_log("setup ogl defaults");
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	glEnable(GL_DEPTH_TEST);

	sw_framebuffer_static_init(); /* TODO: move elsewhere */
	win->_scaler_fb = sw_framebuffer_create(size);

	glClearColor(0.1f, 0.2f, 0.3f, 1.f);
	sw_log("window setup complete");
	return win;
}

void sw_window_run(struct sw_window *window, void (*callback)()) {
	sw_log("run window");
	_tick_window = window;
	_tick_fn = callback;
#ifdef __EMSCRIPTEN__
	emscripten_set_main_loop(_tick, 60, 1);
#else
	while(!glfwWindowShouldClose(window->_window)) {
		_tick();
	}
#endif
}

/* private functions */

void _tick() {
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
	glBindFramebuffer(GL_FRAMEBUFFER, _tick_window->_scaler_fb->_fb);
	_tick_fn();
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
	sw_framebuffer_copy_to_screen(_tick_window->_scaler_fb);
	glfwSwapBuffers(_tick_window->_window);
	glfwPollEvents();
}

void GLAPIENTRY gldebug(
    GLenum source,
    GLenum type,
    GLuint id,
    GLenum severity,
    GLsizei length,
    const char *msg,
    const void *user
) {
	(void)source;
	(void)id;
	(void)severity;
	(void)length;
	(void)user;

	if(type == GL_DEBUG_TYPE_ERROR) {
		sw_error("%s", msg);
	} else {
		sw_log("%s", msg);
	}
}
